package ru.vsu.cs.course3.compiler.parse;

import ru.vsu.cs.course3.compiler.ast.*;

import java.util.*;
import java.util.stream.Collectors;

public class ParseTreeToAstTransformer extends ParserBaseVisitor<Object> {

    @Override
    public LiteralNode visitLiteral(ParserParser.LiteralContext ctx) {
        return new LiteralNode(ctx.getText());
    }

    @Override
    public IdentNode visitIdent(ParserParser.IdentContext ctx) {
        return new IdentNode(ctx.getText());
    }

    @Override
    public ExprNode visitGroupRule(ParserParser.GroupRuleContext ctx) {
        return (ExprNode) visit(ctx.res_);
    }

    @Override
    public BinaryOpNode visitBinOpRule(ParserParser.BinOpRuleContext ctx) {
        ExprNode arg1 = (ExprNode) visit(ctx.arg1_);
        String op = ctx.op_.getText();
        ExprNode arg2 = (ExprNode) visit(ctx.arg2_);
        return new BinaryOpNode(op, arg1, arg2);
    }

    @Override
    public List<ExprNode> visitParams(ParserParser.ParamsContext ctx) {
        @SuppressWarnings("unchecked")
        List<ExprNode> exprs = (List<ExprNode>) (Object) ctx.exprs_.stream().map(this::visit).collect(Collectors.toList());
        return exprs;
    }

    @Override
    public FuncCallNode visitCall(ParserParser.CallContext ctx) {
        IdentNode name = (IdentNode) visit(ctx.name);
        @SuppressWarnings("unchecked")
        List<ExprNode> params = (List<ExprNode>) visit(ctx.params_);
        return new FuncCallNode(name, params);
    }

    @Override
    public AstNode visitAssign(ParserParser.AssignContext ctx) {
        IdentNode name = (IdentNode) visit(ctx.name_);
        ExprNode val = (ExprNode) visit(ctx.val_);
        return new AssignNode(name, val);
    }

    @Override
    public VarsDeclNode visitVarsDecl(ParserParser.VarsDeclContext ctx) {
        IdentNode type = (IdentNode) visit(ctx.type_);
        @SuppressWarnings("unchecked")
        List<AstNode> vars = (List<AstNode>) (Object) ctx.vars_.stream().map(this::visit).collect(Collectors.toList());
        return new VarsDeclNode(type, vars);
    }

    @Override
    public VarsDeclNode visitParamDecl(ParserParser.ParamDeclContext ctx) {
        IdentNode type = (IdentNode) visit(ctx.type_);
        IdentNode name = (IdentNode) visit(ctx.name_);
        return new VarsDeclNode(type, Collections.singletonList(name));
    }

    @Override
    public FuncDeclNode visitFuncDecl(ParserParser.FuncDeclContext ctx) {
        IdentNode type = (IdentNode) visit(ctx.type_);
        IdentNode name = (IdentNode) visit(ctx.name_);
        @SuppressWarnings("unchecked")
        List<VarsDeclNode> params = (List<VarsDeclNode>) (Object) ctx.params_.stream().map(this::visit).collect(Collectors.toList());
        StmtNode body = (StmtNode) visit(ctx.body_);
        return new FuncDeclNode(type, name, params, body);
    }

    @Override
    public StmtNode visitCompoundRule(ParserParser.CompoundRuleContext ctx) {
        return (StmtNode) visit(ctx.res_);
    }

    @Override
    public IfNode visitIfRule(ParserParser.IfRuleContext ctx) {
        ExprNode cond = (ExprNode) visit(ctx.cond_);
        StmtNode thenStmt = (StmtNode) visit(ctx.thenStmt_);
        StmtNode elseStmt = (StmtNode) (ctx.elseStmt_ != null ? visit(ctx.elseStmt_) : null);
        return new IfNode(cond, thenStmt, elseStmt);
    }

    @Override
    public WhileNode visitWhileRule(ParserParser.WhileRuleContext ctx) {
        ExprNode cond = (ExprNode) visit(ctx.cond_);
        StmtNode body = (StmtNode) visit(ctx.body_);
        return new WhileNode(cond, body);
    }

    @Override
    public ForNode visitForRule(ParserParser.ForRuleContext ctx) {
        StmtNode init = ctx.init_ == null ? new StmtListNode() : (StmtNode) visit(ctx.init_);
        ExprNode cond = ctx.cond_ == null ? new LiteralNode("true") : (ExprNode) visit(ctx.cond_);
        StmtNode step = ctx.step_ == null ? new StmtListNode() : (StmtNode) visit(ctx.step_);
        StmtNode body = (StmtNode) visit(ctx.body_);
        return new ForNode(init, cond, step, body);
    }

    @Override
    public StmtNode visitStmtList(ParserParser.StmtListContext ctx) {
        @SuppressWarnings("unchecked")
        List<StmtNode> stmts = (List<StmtNode>) (Object) ctx.stmts_.stream().map(this::visit).collect(Collectors.toList());
        return stmts.size() == 1 ? stmts.stream().findFirst().get() : new StmtListNode(stmts);
    }

    @Override
    public StmtNode visitStmtRule(ParserParser.StmtRuleContext ctx) {
        return (StmtNode) visit(ctx.res_);
    }

    @Override
    public StmtNode visitProgram(ParserParser.ProgramContext ctx) {
        return (StmtNode) visit(ctx.res_);
    }
}
